function crf = crf_learn(goodFeatures,reg,letters,wordidx,pixels,varargin)
% Evaluates the objective (log-likelihood of the data
%
% Usage:
%
%    crf = crf_learn(params,reg,letters,wordidx,pixels)
%
% Learn CRF from the dataset. If M is the number of characters in the dataset, then LETTER is a M x 1 array of letter labels,
% PIXELS is a M x 64 binary matrix of pixel activations for each letter,
% and wordidx is a list of words, each consisting of a sequence of indices into letters/pixels.
%
% Output is crf structure
% crf.paramsNode - 64 x 26 matrix
% crf.paramsEdge - 26 x 26 matrix
% crf.infer - inference routine, computes node/edge marginals given new
% feature observations using learned parameters

%Get number of parameters, indices of good features %

npNode  =    nnz(goodFeatures.nodes);
npEdge  =    nnz(goodFeatures.edges);
np      =    npNode+npEdge;

%You can change this if running to convergence takes too long
niter   =   100; %Number of iterations of likelihood optimization algorithm
nfeval  =   100; %Number of objective evaluations of likelihood optimization algorithm

%   Optimize likelihood wrt parameters  %

%Option to initialize with precomputed sensible parameters
if(nargin>5)
    params  =   varargin{1};
else
    params  =   randn(np,1);
end

params  =    minFunc(@EvalLikl,params,struct('MaxIter',niter,'MaxFunEvals',nfeval));

%   Store Results in Structure  %

crf.params                          =    params;
crf.paramsNode                      =   zeros(64,26);
crf.paramsNode(goodFeatures.nodes)  =   params(1:npNode);
crf.paramsEdge                      =   zeros(26);
crf.paramsEdge(goodFeatures.edges)  =   params(npNode+(1:npEdge));

crf.infer                           =   @(obs) crf_infer(params,obs);




%Loops over data, computes likelihood and its gradient averaged over
%dataset, adds regularization
    function [logLikl,gradient]=EvalLikl(params)
        
        logLikl         =   0;
        gradient        =   zeros(size(params));
        
        %Loop over each word in the dataset
        for it=1:length(wordidx)
            
            %   Construct feature vectors (for node potentials)
            xs           =   pixels(wordidx{it},:)';
            ys           =   letters(wordidx{it});
            
            %Infer Likelihood and Gradient
            [lik,grad] =   inferlikelihood(params,xs,ys);
            
            %Add to overall likelihood/gradient
            logLikl     =   logLikl+lik;
            gradient    =   gradient+grad;
        end
        
        %Scale Objective and Gradient by number of samples
        logLikl =   logLikl/length(letters);
        gradient=   gradient/length(letters);
        
        %Negate likelihood/gradient for minimization
        
        logLikl     =   -logLikl;
        gradient    =   -gradient;
        
        %Add Tikhonov regularization
        
        logLikl     =   logLikl+params'*params*reg/2;
        gradient    =   gradient+reg*params;
        
        
        
    end

% YOU need to implement the function inferlikelihood,crf_infer:
% Evaluate likelihood for given sequence of
% features (obs) and character labels (labels),
% CRF parameters

%To be implemented%
%To be implemented by the student
    function [l,g]=inferlikelihood(param,obs,labels)
        T           =   size(obs,2);
        paramsNode  =   zeros(64,26);
        paramsNode(goodFeatures.nodes)    =   param(1:npNode);
        
        paramsEdge  =   zeros(26);
        paramsEdge(goodFeatures.edges)   =   param(npNode+(1:npEdge));
        
        %Perform inference in CRF
        %Call to crf_infer
                
        l   =   0;
        g   =   zeros(size(param));
    end

%To be implemented by student
    function [nodeMarginals,edgeMarginals,logZ]=crf_infer(param,obs)
        %Construct Node and Edge potentials 
        %nodePots: 26 x T
        %edgePots: 26 x 26 - fixed in time
        
        [nodePots,edgePots]    =   makePotentials(param,obs);
        T                      =   size(obs,2);
        
        
        %perform inference, compute nodeMarginals - 26 x T, edgeMarginals -
        %26 x 26 x (T-1) and logZ - 1x1 - log partition function
    end

    function [nodePots,edgePots]=makePotentials(param,obs)
        
        T           =   size(obs,2);
        
        paramsNode  =   zeros(64,26);
        paramsNode(goodFeatures.nodes)   =   param(1:npNode);
        
        paramsEdge  =   zeros(26);
        paramsEdge(goodFeatures.edges)   =   param(npNode+(1:npEdge));
        
        nodePots    =   zeros(26,T);
        for it=1:T
            for jt=1:26
                nodePots(jt,it)       =   paramsNode(:,jt)'*obs(:,it);
            end
        end
        
        edgePots    =   paramsEdge;
        
    end
    %Useful function to perform computations in log domain.
    % M - matrix
    % i = 1 or 2
    % returns log(sum(exp(M),i)), but carefully computed in log domain to
    % avoid numerical issues
    function res=logsumExp(M,i)
        scl =   max(M,[],i);
        M   =   bsxfun(@minus,M,scl);
        res =   log(sum(exp(M),i))+scl;
        res =   res(:);
    end
end

